""" Module that holds the ReportGeneratorHandler class

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ReportGenerator
  :end-before: ##END
  :dedent: 2
  :caption: ReportGenerator options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import datetime

from DIRAC import S_OK, S_ERROR, rootPath, gConfig, gLogger
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities import Time
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.AccountingSystem.DB.MultiAccountingDB import MultiAccountingDB
from DIRAC.Core.Utilities.Plotting import gDataCache
from DIRAC.AccountingSystem.private.MainReporter import MainReporter
from DIRAC.AccountingSystem.private.DBUtils import DBUtils
from DIRAC.AccountingSystem.private.Policies import gPoliciesList
from DIRAC.Core.Utilities.Plotting.Plots import generateErrorMessagePlot
from DIRAC.Core.Utilities.Plotting.FileCoding import extractRequestFromFileId
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.RequestHandler import RequestHandler


class ReportGeneratorHandler(RequestHandler):
    """DIRAC service class to retrieve information from the AccountingDB"""

    __acDB = None
    __reportRequestDict = {
        "typeName": six.string_types,
        "reportName": six.string_types,
        "startTime": (datetime.datetime, datetime.date),
        "endTime": (datetime.datetime, datetime.date),
        "condDict": dict,
        "grouping": six.string_types,
        "extraArgs": dict,
    }

    @classmethod
    def initializeHandler(cls, serviceInfo):
        multiPath = PathFinder.getDatabaseSection("Accounting/MultiDB")
        cls.__acDB = MultiAccountingDB(multiPath, readOnly=True)
        # Get data location
        reportSection = serviceInfo["serviceSectionPath"]
        dataPath = gConfig.getValue("%s/DataLocation" % reportSection, "data/accountingGraphs")
        dataPath = dataPath.strip()
        if "/" != dataPath[0]:
            dataPath = os.path.realpath("%s/%s" % (gConfig.getValue("/LocalSite/InstancePath", rootPath), dataPath))
        gLogger.info("Data will be written into %s" % dataPath)
        mkDir(dataPath)
        try:
            testFile = "%s/acc.jarl.test" % dataPath
            with open(testFile, "w"):
                pass
            os.unlink(testFile)
        except IOError:
            gLogger.fatal("Can't write to %s" % dataPath)
            return S_ERROR("Data location is not writable")
        gDataCache.setGraphsLocation(dataPath)
        gMonitor.registerActivity("plotsDrawn", "Drawn plot images", "Accounting reports", "plots", gMonitor.OP_SUM)
        gMonitor.registerActivity(
            "reportsRequested", "Generated reports", "Accounting reports", "reports", gMonitor.OP_SUM
        )
        return S_OK()

    def __checkPlotRequest(self, reportRequest):
        # If extraArgs is not there add it
        if "extraArgs" not in reportRequest:
            reportRequest["extraArgs"] = {}
        if not isinstance(reportRequest["extraArgs"], self.__reportRequestDict["extraArgs"]):
            return S_ERROR("Extra args has to be of type %s" % self.__reportRequestDict["extraArgs"])
        reportRequestExtra = reportRequest["extraArgs"]
        # Check sliding plots
        if "lastSeconds" in reportRequestExtra:
            try:
                lastSeconds = int(reportRequestExtra["lastSeconds"])
            except ValueError:
                gLogger.error("lastSeconds key must be a number")
                return S_ERROR("Value Error")
            if lastSeconds < 3600:
                return S_ERROR("lastSeconds must be more than 3600")
            now = Time.dateTime()
            reportRequest["endTime"] = now
            reportRequest["startTime"] = now - datetime.timedelta(seconds=lastSeconds)
        else:
            # if enddate is not there, just set it to now
            if not reportRequest.get("endTime", False):
                reportRequest["endTime"] = Time.dateTime()
        # Check keys
        for key in self.__reportRequestDict:
            if key not in reportRequest:
                return S_ERROR("Missing mandatory field %s in plot reques" % key)

            if not isinstance(reportRequest[key], self.__reportRequestDict[key]):
                return S_ERROR(
                    "Type mismatch for field %s (%s), required one of %s"
                    % (key, str(type(reportRequest[key])), str(self.__reportRequestDict[key]))
                )
            if key in ("startTime", "endTime"):
                reportRequest[key] = int(Time.toEpoch(reportRequest[key]))

        return S_OK(reportRequest)

    types_generatePlot = [dict]

    def export_generatePlot(self, reportRequest):
        """
        Plot a accounting

        Arguments:
          - viewName : Name of view (easy!)
          - startTime
          - endTime
          - argsDict : Arguments to the view.
          - grouping
          - extraArgs
        """
        retVal = self.__checkPlotRequest(reportRequest)
        if not retVal["OK"]:
            return retVal
        reporter = MainReporter(self.__acDB, self.serviceInfoDict["clientSetup"])
        gMonitor.addMark("plotsDrawn")
        reportRequest["generatePlot"] = True
        return reporter.generate(reportRequest, self.getRemoteCredentials())

    types_getReport = [dict]

    def export_getReport(self, reportRequest):
        """
        Plot a accounting

        Arguments:
          - viewName : Name of view (easy!)
          - startTime
          - endTime
          - argsDict : Arguments to the view.
          - grouping
          - extraArgs
        """
        retVal = self.__checkPlotRequest(reportRequest)
        if not retVal["OK"]:
            return retVal
        reporter = MainReporter(self.__acDB, self.serviceInfoDict["clientSetup"])
        gMonitor.addMark("reportsRequested")
        reportRequest["generatePlot"] = False
        return reporter.generate(reportRequest, self.getRemoteCredentials())

    types_listReports = [six.string_types]

    def export_listReports(self, typeName):
        """
        List all available plots

        Arguments:
          - none
        """
        reporter = MainReporter(self.__acDB, self.serviceInfoDict["clientSetup"])
        return reporter.list(typeName)

    types_listUniqueKeyValues = [six.string_types]

    def export_listUniqueKeyValues(self, typeName):
        """
        List all values for all keys in a type

        Arguments:
          - none
        """
        dbUtils = DBUtils(self.__acDB, self.serviceInfoDict["clientSetup"])
        credDict = self.getRemoteCredentials()
        if typeName in gPoliciesList:
            policyFilter = gPoliciesList[typeName]
            filterCond = policyFilter.getListingConditions(credDict)
        else:
            policyFilter = gPoliciesList["Null"]
            filterCond = {}
        retVal = dbUtils.getKeyValues(typeName, filterCond)
        if not policyFilter or not retVal["OK"]:
            return retVal
        return policyFilter.filterListingValues(credDict, retVal["Value"])

    def __generatePlotFromFileId(self, fileId):
        result = extractRequestFromFileId(fileId)
        if not result["OK"]:
            return result
        plotRequest = result["Value"]
        gLogger.info("Generating the plots..")
        result = self.export_generatePlot(plotRequest)
        if not result["OK"]:
            gLogger.error("Error while generating the plots", result["Message"])
            return result
        fileToReturn = "plot"
        if "extraArgs" in plotRequest:
            extraArgs = plotRequest["extraArgs"]
            if "thumbnail" in extraArgs and extraArgs["thumbnail"]:
                fileToReturn = "thumbnail"
        gLogger.info("Returning %s file: %s " % (fileToReturn, result["Value"][fileToReturn]))
        return S_OK(result["Value"][fileToReturn])

    def __sendErrorAsImg(self, msgText, fileHelper):
        retVal = generateErrorMessagePlot(msgText)
        retVal = fileHelper.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        fileHelper.sendEOF()
        return S_OK()

    def transfer_toClient(self, fileId, token, fileHelper):
        """
        Get graphs data
        """
        # First check if we've got to generate the plot
        if len(fileId) > 5 and fileId[1] == ":":
            gLogger.info("Seems the file request is a plot generation request!")
            # Seems a request for a plot!
            try:
                result = self.__generatePlotFromFileId(fileId)
            except Exception as e:
                gLogger.exception("Exception while generating plot")
                result = S_ERROR("Error while generating plot: %s" % str(e))
            if not result["OK"]:
                self.__sendErrorAsImg(result["Message"], fileHelper)
                fileHelper.sendEOF()
                return result
            fileId = result["Value"]
        retVal = gDataCache.getPlotData(fileId)
        if not retVal["OK"]:
            self.__sendErrorAsImg(retVal["Message"], fileHelper)
            return retVal
        retVal = fileHelper.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        fileHelper.sendEOF()
        return S_OK()



""" EmailAction

  This action writes all the necessary data to a cache file ( cache.db ) that
  will be used later by the EmailAgent in order to send the emails for each site.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import S_ERROR, S_OK
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE
from DIRAC.ResourceStatusSystem.PolicySystem.Actions.BaseAction import BaseAction


class EmailAction(BaseAction):
    def __init__(self, name, decisionParams, enforcementResult, singlePolicyResults, clients=None):

        super(EmailAction, self).__init__(name, decisionParams, enforcementResult, singlePolicyResults, clients)

        if clients is not None and "ResourceStatusClient" in clients:
            self.rsClient = clients["ResourceStatusClient"]
        else:
            self.rsClient = ResourceStatusClient()

    def run(self):
        """Checks it has the parameters it needs and writes the date to a cache file."""
        # Minor security checks

        element = self.decisionParams["element"]
        if element is None:
            return S_ERROR("element should not be None")

        name = self.decisionParams["name"]
        if name is None:
            return S_ERROR("name should not be None")

        statusType = self.decisionParams["statusType"]
        if statusType is None:
            return S_ERROR("statusType should not be None")

        previousStatus = self.decisionParams["status"]
        if previousStatus is None:
            return S_ERROR("status should not be None")

        status = self.enforcementResult["Status"]
        if status is None:
            return S_ERROR("status should not be None")

        reason = self.enforcementResult["Reason"]
        if reason is None:
            return S_ERROR("reason should not be None")

        if self.decisionParams["element"] == "Site":
            siteName = self.decisionParams["name"]
        else:
            elementType = self.decisionParams["elementType"]

            if elementType == "StorageElement":
                siteName = getSitesForSE(name)
            elif elementType == "ComputingElement":
                res = getCESiteMapping(name)
                if not res["OK"]:
                    return res
                siteName = S_OK(res["Value"][name])
            else:
                siteName = {"OK": True, "Value": "Unassigned"}

            if not siteName["OK"]:
                self.log.error("Resource %s does not exist at any site: %s" % (name, siteName["Message"]))
                siteName = "Unassigned Resources"
            elif not siteName["Value"]:
                siteName = "Unassigned Resources"
            else:
                siteName = (
                    siteName["Value"] if isinstance(siteName["Value"], six.string_types) else siteName["Value"][0]
                )

        # create record for insertion
        recordDict = {}
        recordDict["SiteName"] = siteName
        recordDict["ResourceName"] = name
        recordDict["Status"] = status
        recordDict["PreviousStatus"] = previousStatus
        recordDict["StatusType"] = statusType

        return self.rsClient.insert("ResourceStatusCache", recordDict)



"""
.. module:: Pfn

:synopsis: ProcessPool and related classes

ProcessPool

ProcessPool creates a pool of worker subprocesses to handle a queue of tasks
much like the producers/consumers paradigm. Users just need to fill the queue
with tasks to be executed and worker tasks will execute them.

To construct ProcessPool one first should call its constructor::

  pool = ProcessPool( minSize, maxSize, maxQueuedRequests )

where parameters are:

  :param int minSize: at least <minSize> workers will be alive all the time
  :param int maxSize: no more than <maxSize> workers will be alive all the time
  :param int maxQueuedRequests: size for request waiting in a queue to be executed

In case another request is added to the full queue, the execution will
lock until another request is taken out. The ProcessPool will automatically increase and
decrease the pool of workers as needed, of course not exceeding above limits.

To add a task to the queue one should execute:::

  pool.createAndQueueTask( funcDef,
                           args = ( arg1, arg2, ... ),
                           kwargs = { "kwarg1" : value1, "kwarg2" : value2 },
                           callback = callbackDef,
                           exceptionCallback = exceptionCallBackDef )

or alternatively by using ProcessTask instance:::

  task = ProcessTask( funcDef,
                      args = ( arg1, arg2, ... )
                      kwargs = { "kwarg1" : value1, .. },
                      callback = callbackDef,
                      exceptionCallback = exceptionCallbackDef )

  pool.queueTask( task )

where parameters are:

  :param funcDef: callable by object definition (function, lambda, class with __call__ slot defined
  :param list args: argument list
  :param dict kwargs: keyword arguments dictionary
  :param callback: callback function definition
  :param exceptionCallback: exception callback function definition

The callback, exceptionCallbaks and the parameters are all optional. Once task has been added to the pool,
it will be executed as soon as possible. Worker subprocesses automatically return the return value of the task.
To obtain those results one has to execute::

  pool.processRequests()

This method will process the existing return values of the task, even if the task does not return
anything. This method has to be called to clean the result queues. To wait until all the requests are finished
and process their result call::

  pool.processAllRequests()

This function will block until all requests are finished and their result values have been processed.

It is also possible to set the ProcessPool in daemon mode, in which all results are automatically
processed as soon they are available, just after finalization of task execution. To enable this mode one
has to call::

  pool.daemonize()

Callback functions

There are two types of callbacks that can be executed for each tasks: exception callback function and
results callback function. The first one is executed when unhandled exception has been raised during
task processing, and hence no task results are available, otherwise the execution of second callback type
is performed.

The callbacks could be attached in a two places:

  - directly in ProcessTask, in that case those have to be shelvable/picklable, so they should be defined as
    global functions with the signature :callback( task, taskResult ): where :task: is a :ProcessTask:
    reference and :taskResult: is whatever task callable it returning for results callback and
    :exceptionCallback( task, exc_info): where exc_info is a
    :S_ERROR( "Exception": { "Value" : exceptionName, "Exc_info" : exceptionInfo ):

  - in ProcessPool, in that case there is no limitation on the function type, except the signature, which
    should follow :callback( task ): or :exceptionCallback( task ):, as those callbacks definitions
    are not put into the queues

The first types of callbacks could be used in case various callable objects are put into the ProcessPool,
so you probably want to handle them differently depending on their results, while the second types are for
executing same type of callables in subprocesses and  hence you are expecting the same type of results
everywhere.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import errno
import inspect
import multiprocessing
import os
import signal
import sys
import threading
import time

import six
from six.moves import queue as Queue

try:
    from DIRAC.FrameworkSystem.Client.Logger import gLogger
except ImportError:
    gLogger = None

try:
    from DIRAC.Core.Utilities.LockRing import LockRing
except ImportError:
    LockRing = None

try:
    from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
except ImportError:

    def S_OK(val=""):
        """dummy S_OK"""
        return {"OK": True, "Value": val}

    def S_ERROR(mess):
        """dummy S_ERROR"""
        return {"OK": False, "Message": mess}


sLog = gLogger.getSubLogger(__name__)


class WorkingProcess(multiprocessing.Process):
    """
    .. class:: WorkingProcess

    WorkingProcess is a class that represents activity that runs in a separate process.

    It is running main thread (process) in daemon mode, reading tasks from :pendingQueue:, executing
    them and pushing back tasks with results to the :resultsQueue:. If task has got a timeout value
    defined a separate threading.Timer thread is started killing execution (and destroying worker)
    after :ProcessTask.__timeOut: seconds.

    Main execution could also terminate in a few different ways:

      * on every failed read attempt (from empty  :pendingQueue:), the  idle loop counter is increased,
        worker is terminated when counter is reaching a value of 10;
      * when stopEvent is set (so ProcessPool is in draining mode),
      * when parent process PID is set to 1 (init process, parent process with ProcessPool is dead).

    """

    def __init__(self, pendingQueue, resultsQueue, stopEvent, keepRunning):
        """c'tor

        :param self: self reference
        :param pendingQueue: queue storing ProcessTask before exection
        :type pendingQueue: multiprocessing.Queue
        :param resultsQueue: queue storing callbacks and exceptionCallbacks
        :type resultsQueue: multiprocessing.Queue
        :param stopEvent: event to stop processing
        :type stopEvent: multiprocessing.Event
        """
        multiprocessing.Process.__init__(self)
        # daemonize
        self.daemon = True
        # flag to see if task is being treated
        self.__working = multiprocessing.Value("i", 0)
        # task counter
        self.__taskCounter = multiprocessing.Value("i", 0)
        # task queue
        self.__pendingQueue = pendingQueue
        # results queue
        self.__resultsQueue = resultsQueue
        # stop event
        self.__stopEvent = stopEvent
        # keep process running until stop event
        self.__keepRunning = keepRunning
        # placeholder for watchdog thread
        self.__watchdogThread = None
        # placeholder for process thread
        self.__processThread = None
        # placeholder for current task
        self.task = None
        # start yourself at least
        self.start()

    def __watchdog(self):
        """
        Watchdog thread target

        Terminating/killing WorkingProcess when parent process is dead

        :param self: self reference
        """
        while True:
            # parent is dead,  commit suicide
            if os.getppid() == 1:
                os.kill(self.pid, signal.SIGTERM)
                # wait for half a minute and if worker is still alive use REAL silencer
                time.sleep(30)
                # now you're dead
                os.kill(self.pid, signal.SIGKILL)
            # wake me up in 5 seconds
            time.sleep(5)

    def isWorking(self):
        """
        Check if process is being executed

        :param self: self reference
        """
        return self.__working.value == 1

    def taskProcessed(self):
        """
        Tell how many tasks have been processed so far

        :param self: self reference
        """
        return self.__taskCounter

    def __processTask(self):
        """
        processThread target

        :param self: self reference
        """
        if self.task:
            self.task.process()

    def run(self):
        """
        Task execution

        Reads and executes ProcessTask :task: out of pending queue and then pushes it
        to the results queue for callback execution.

        :param self: self reference
        """
        # start watchdog thread
        self.__watchdogThread = threading.Thread(target=self.__watchdog)
        self.__watchdogThread.daemon = True
        self.__watchdogThread.start()

        if LockRing:
            # Reset all locks
            lr = LockRing()
            lr._openAll()
            lr._setAllEvents()

        # zero processed task counter
        taskCounter = 0
        # zero idle loop counter
        idleLoopCount = 0

        # main loop
        while True:

            # draining, stopEvent is set, exiting
            if self.__stopEvent.is_set():
                return

            # clear task
            self.task = None

            # read from queue
            try:
                task = self.__pendingQueue.get(block=True, timeout=10)
            except Queue.Empty:
                # idle loop?
                idleLoopCount += 1
                # 10th idle loop - exit, nothing to do
                if idleLoopCount == 10 and not self.__keepRunning:
                    return
                continue

            # toggle __working flag
            self.__working.value = 1
            # save task
            self.task = task
            # reset idle loop counter
            idleLoopCount = 0

            # process task in a separate thread
            self.__processThread = threading.Thread(target=self.__processTask)
            self.__processThread.start()

            timeout = False
            noResults = False
            # join processThread with or without timeout
            if self.task.getTimeOut():
                self.__processThread.join(self.task.getTimeOut() + 10)
            else:
                self.__processThread.join()

            # processThread is still alive? stop it!
            if self.__processThread.is_alive():
                if six.PY2:
                    self.__processThread._Thread__stop()  # pylint: disable=no-member
                self.task.setResult(S_ERROR(errno.ETIME, "Timed out"))
                timeout = True
            # if the task finished with no results, something bad happened, e.g.
            # undetected timeout
            if not self.task.taskResults() and not self.task.taskException():
                self.task.setResult(S_ERROR("Task produced no results"))
                noResults = True

            # check results and callbacks presence, put task to results queue
            if self.task.hasCallback() or self.task.hasPoolCallback():
                self.__resultsQueue.put(task)
            if timeout or noResults:
                # The task execution timed out, stop the process to prevent it from running
                # in the background
                time.sleep(1)
                os.kill(self.pid, signal.SIGKILL)
                return
            # increase task counter
            taskCounter += 1
            self.__taskCounter = taskCounter
            # toggle __working flag
            self.__working.value = 0


class ProcessTask(object):
    """Defines task to be executed in WorkingProcess together with its callbacks."""

    # taskID
    taskID = 0

    def __init__(
        self,
        taskFunction,
        args=None,
        kwargs=None,
        taskID=None,
        callback=None,
        exceptionCallback=None,
        usePoolCallbacks=False,
        timeOut=0,
    ):
        """c'tor

        :warning: taskFunction has to be callable: it could be a function, lambda OR a class with
          __call__ operator defined. But be carefull with interpretation of args and kwargs, as they
          are passed to different places in above cases:

        1. for functions or lambdas args and kwargs are just treated as function parameters

        2. for callable classess (say MyTask) args and kwargs are passed to class contructor
           (MyTask.__init__) and MyTask.__call__ should be a method without parameters, i.e.
           MyTask definition should be::

            class MyTask:
              def __init__( self, *args, **kwargs ):
                ...
              def __call__( self ):
                ...

        :warning: depending on :timeOut: value, taskFunction execution can be forcefully terminated
          using SIGALRM after :timeOut: seconds spent, :timeOut: equal to zero means there is no any
          time out at all, except those during :ProcessPool: finalization

        :param self: self reference
        :param mixed taskFunction: definition of callable object to be executed in this task
        :param tuple args: non-keyword arguments
        :param dict kwargs: keyword arguments
        :param int taskID: task id, if not set,
        :param int timeOut: estimated time to execute taskFunction in seconds (default = 0, no timeOut at all)
        :param mixed callback: result callback function
        :param mixed exceptionCallback: callback function to be fired upon exception in taskFunction
        """
        self.__taskFunction = taskFunction
        self.__taskArgs = args or []
        self.__taskKwArgs = kwargs or {}
        self.__taskID = taskID
        self.__resultCallback = callback
        self.__exceptionCallback = exceptionCallback
        self.__timeOut = 0
        # set time out
        self.setTimeOut(timeOut)
        self.__done = False
        self.__exceptionRaised = False
        self.__taskException = None
        self.__taskResult = None
        self.__usePoolCallbacks = usePoolCallbacks

    def taskResults(self):
        """
        Get task results

        :param self: self reference
        """
        return self.__taskResult

    def taskException(self):
        """
        Get task exception

        :param self: self reference
        """
        return self.__taskException

    def enablePoolCallbacks(self):
        """
        (re)enable use of ProcessPool callbacks
        """
        self.__usePoolCallbacks = True

    def disablePoolCallbacks(self):
        """
        Disable execution of ProcessPool callbacks
        """
        self.__usePoolCallbacks = False

    def usePoolCallbacks(self):
        """
        Check if results should be processed by callbacks defined in the :ProcessPool:

        :param self: self reference
        """
        return self.__usePoolCallbacks

    def hasPoolCallback(self):
        """
        Check if asked to execute :ProcessPool: callbacks

        :param self: self reference
        """
        return self.__usePoolCallbacks

    def setTimeOut(self, timeOut):
        """
        Set time out (in seconds)

        :param self: selt reference
        :param int timeOut: new time out value
        """
        try:
            self.__timeOut = int(timeOut)
            return S_OK(self.__timeOut)
        except (TypeError, ValueError) as error:
            return S_ERROR(str(error))

    def getTimeOut(self):
        """
        Get timeOut value

        :param self: self reference
        """
        return self.__timeOut

    def hasTimeOutSet(self):
        """
        Check if timeout is set

        :param self: self reference
        """
        return bool(self.__timeOut != 0)

    def getTaskID(self):
        """
        TaskID getter

        :param self: self reference
        """
        return self.__taskID

    def hasCallback(self):
        """
        Callback existence checking

        :param self: self reference
        :return: True if callback or exceptionCallback has been defined, False otherwise
        """
        return self.__resultCallback or self.__exceptionCallback or self.__usePoolCallbacks

    def exceptionRaised(self):
        """
        Flag to determine exception in process

        :param self: self reference
        """
        return self.__exceptionRaised

    def doExceptionCallback(self):
        """
        Execute exceptionCallback

        :param self: self reference
        """
        if self.__done and self.__exceptionRaised and self.__exceptionCallback:
            self.__exceptionCallback(self, self.__taskException)

    def doCallback(self):
        """
        Execute result callback function

        :param self: self reference
        """
        if self.__done and not self.__exceptionRaised and self.__resultCallback:
            self.__resultCallback(self, self.__taskResult)

    def setResult(self, result):
        """
        Set taskResult to result
        """
        self.__taskResult = result

    def process(self):
        """
        Execute task

        :param self: self reference
        """
        self.__done = True
        try:
            # it's a function?
            if inspect.isfunction(self.__taskFunction):
                self.__taskResult = self.__taskFunction(*self.__taskArgs, **self.__taskKwArgs)
            # or a class?
            elif inspect.isclass(self.__taskFunction):
                # create new instance
                taskObj = self.__taskFunction(*self.__taskArgs, **self.__taskKwArgs)
                # ## check if it is callable, raise TypeError if not
                if not callable(taskObj):
                    raise TypeError("__call__ operator not defined not in %s class" % taskObj.__class__.__name__)
                # ## call it at least
                self.__taskResult = taskObj()
        except Exception as x:
            self.__exceptionRaised = True
            if gLogger:
                gLogger.exception("Exception in process of pool")
            if self.__exceptionCallback or self.usePoolCallbacks():
                retDict = S_ERROR("Exception")
                retDict["Value"] = str(x)
                retDict["Exc_info"] = sys.exc_info()[1]
                self.__taskException = retDict


class ProcessPool(object):
    """
    .. class:: ProcessPool

    ProcessPool

    This class is managing multiprocessing execution of tasks (:ProcessTask: instances) in a separate
    sub-processes (:WorkingProcess:).

    Pool depth

    The :ProcessPool: is keeping required number of active workers all the time: slave workers are only created
    when pendingQueue is being filled with tasks, not exceeding defined min and max limits. When pendingQueue is
    empty, active workers will be cleaned up by themselves, as each worker has got built in
    self-destroy mechanism after 10 idle loops.

    Processing and communication

    The communication between :ProcessPool: instance and slaves is performed using two :multiprocessing.Queues:

      * pendingQueue, used to push tasks to the workers,
      * resultsQueue for revert direction;

    and one :multiprocessing.Event: instance (stopEvent), which is working as a fuse to destroy idle workers
    in a clean manner.

    Processing of task begins with pushing it into :pendingQueue: using :ProcessPool.queueTask: or
    :ProcessPool.createAndQueueTask:. Every time new task is queued, :ProcessPool: is checking existance of
    active and idle workers and spawning new ones when required. The task is then read and processed on worker
    side. If results are ready and callback functions are defined, task is put back to the resultsQueue and it is
    ready to be picked up by ProcessPool again. To perform this last step one has to call :ProcessPool.processResults:,
    or alternatively ask for daemon mode processing, when this function is called again and again in
    separate background thread.

    Finalisation

    Finalization for task processing is done in several steps:

      * if pool is working in daemon mode, background result processing thread is joined and stopped
      * :pendingQueue: is emptied by :ProcessPool.processAllResults: function, all enqueued tasks are executed
      * :stopEvent: is set, so all idle workers are exiting immediately
      * non-hanging workers are joined and terminated politelty
      * the rest of workers, if any, are forcefully retained by signals: first by SIGTERM, and if is doesn't work
        by SIGKILL

    :warn: Be carefull and choose wisely :timeout: argument to :ProcessPool.finalize:. Too short time period can
           cause that all workers will be killed.

    """

    def __init__(
        self,
        minSize=2,
        maxSize=0,
        maxQueuedRequests=10,
        strictLimits=True,
        poolCallback=None,
        poolExceptionCallback=None,
        keepProcessesRunning=True,
    ):
        """c'tor

        :param self: self reference
        :param int minSize: minimal number of simultaniously executed tasks
        :param int maxSize: maximal number of simultaniously executed tasks
        :param int maxQueueRequests: size of pending tasks queue
        :param bool strictLimits: flag to workers overcommitment
        :param callable poolCallbak: results callback
        :param callable poolExceptionCallback: exception callback
        """
        # min workers
        self.__minSize = max(1, minSize)
        # max workers
        self.__maxSize = max(self.__minSize, maxSize)
        # queue size
        self.__maxQueuedRequests = maxQueuedRequests
        # flag to worker overcommit
        self.__strictLimits = strictLimits

        # pool results callback
        self.__poolCallback = poolCallback
        # pool exception callback
        self.__poolExceptionCallback = poolExceptionCallback

        # pending queue
        self.__pendingQueue = multiprocessing.Queue(self.__maxQueuedRequests)
        # results queue
        self.__resultsQueue = multiprocessing.Queue(0)
        # stop event
        self.__stopEvent = multiprocessing.Event()
        # keep processes running flag
        self.__keepRunning = keepProcessesRunning
        # lock
        self.__prListLock = threading.Lock()

        # workers dict
        self.__workersDict = {}

        # flag to trigger workers draining
        self.__draining = False
        # placeholder for daemon results processing
        self.__daemonProcess = False

        # create initial workers
        self.__spawnNeededWorkingProcesses()

    def stopProcessing(self, timeout=10):
        """
        Case fire

        :param self: self reference
        """
        self.finalize(timeout)

    def startProcessing(self):
        """
        Restart processing again

        :param self: self reference
        """
        self.__draining = False
        self.__stopEvent.clear()
        self.daemonize()

    def setPoolCallback(self, callback):
        """
        Set ProcessPool callback function

        :param self: self reference
        :param callable callback: callback function
        """
        if callable(callback):
            self.__poolCallback = callback

    def setPoolExceptionCallback(self, exceptionCallback):
        """
        Set ProcessPool exception callback function

        :param self: self refernce
        :param callable exceptionCallback: exsception callback function
        """
        if callable(exceptionCallback):
            self.__poolExceptionCallback = exceptionCallback

    def getMaxSize(self):
        """
        MaxSize getter

        :param self: self reference
        """
        return self.__maxSize

    def getMinSize(self):
        """
        MinSize getter

        :param self: self reference
        """
        return self.__minSize

    def getNumWorkingProcesses(self):
        """
        Count processes currently being executed

        :param self: self reference
        """
        counter = 0
        self.__prListLock.acquire()
        try:
            counter = len([pid for pid, worker in self.__workersDict.items() if worker.isWorking()])
        finally:
            self.__prListLock.release()
        return counter

    def getNumIdleProcesses(self):
        """
        Count processes being idle

        :param self: self reference
        """
        counter = 0
        self.__prListLock.acquire()
        try:
            counter = len([pid for pid, worker in self.__workersDict.items() if not worker.isWorking()])
        finally:
            self.__prListLock.release()
        return counter

    def getFreeSlots(self):
        """get number of free slots available for workers

        :param self: self reference
        """
        return max(0, self.__maxSize - self.getNumWorkingProcesses())

    def __spawnWorkingProcess(self):
        """
        Create new process

        :param self: self reference
        """
        self.__prListLock.acquire()
        try:
            worker = WorkingProcess(self.__pendingQueue, self.__resultsQueue, self.__stopEvent, self.__keepRunning)
            while worker.pid is None:
                time.sleep(0.1)
            self.__workersDict[worker.pid] = worker
        finally:
            self.__prListLock.release()

    def __cleanDeadProcesses(self):
        """
        Delete references of dead workingProcesses from ProcessPool.__workingProcessList
        """
        # check wounded processes
        self.__prListLock.acquire()
        try:
            for pid, worker in list(self.__workersDict.items()):
                if not worker.is_alive():
                    del self.__workersDict[pid]
        finally:
            self.__prListLock.release()

    def __spawnNeededWorkingProcesses(self):
        """
        Create N working process (at least self.__minSize, but no more
        than self.__maxSize)

        :param self: self reference
        """
        self.__cleanDeadProcesses()
        # if we're draining do not spawn new workers
        if self.__draining or self.__stopEvent.is_set():
            return

        while len(self.__workersDict) < self.__minSize:
            if self.__draining or self.__stopEvent.is_set():
                return
            self.__spawnWorkingProcess()

        while self.hasPendingTasks() and self.getNumIdleProcesses() == 0 and len(self.__workersDict) < self.__maxSize:
            if self.__draining or self.__stopEvent.is_set():
                return
            self.__spawnWorkingProcess()
            time.sleep(0.1)

    def queueTask(self, task, blocking=True, usePoolCallbacks=False):
        """
        Enqueue new task into pending queue

        :param self: self reference
        :param ProcessTask task: new task to execute
        :param bool blocking: flag to block if necessary and new empty slot is available (default = block)
        :param bool usePoolCallbacks: flag to trigger execution of pool callbacks (default = don't execute)
        """
        if not isinstance(task, ProcessTask):
            raise TypeError("Tasks added to the process pool must be ProcessTask instances")
        if usePoolCallbacks and (self.__poolCallback or self.__poolExceptionCallback):
            task.enablePoolCallbacks()

        self.__prListLock.acquire()
        try:
            self.__pendingQueue.put(task, block=blocking)
        except Queue.Full:
            self.__prListLock.release()
            return S_ERROR("Queue is full")
        finally:
            self.__prListLock.release()

        self.__spawnNeededWorkingProcesses()
        # throttle a bit to allow task state propagation
        time.sleep(0.1)
        return S_OK()

    def createAndQueueTask(
        self,
        taskFunction,
        args=None,
        kwargs=None,
        taskID=None,
        callback=None,
        exceptionCallback=None,
        blocking=True,
        usePoolCallbacks=False,
        timeOut=0,
    ):
        """
        Create new processTask and enqueue it in pending task queue

        :param self: self reference
        :param mixed taskFunction: callable object definition (FunctionType, LambdaType, callable class)
        :param tuple args: non-keyword arguments passed to taskFunction c'tor
        :param dict kwargs: keyword arguments passed to taskFunction c'tor
        :param int taskID: task Id
        :param mixed callback: callback handler, callable object executed after task's execution
        :param mixed exceptionCallback: callback handler executed if testFunction had raised an exception
        :param bool blocking: flag to block queue if necessary until free slot is available
        :param bool usePoolCallbacks: fire execution of pool defined callbacks after task callbacks
        :param int timeOut: time you want to spend executing :taskFunction:
        """
        task = ProcessTask(taskFunction, args, kwargs, taskID, callback, exceptionCallback, usePoolCallbacks, timeOut)
        return self.queueTask(task, blocking)

    def hasPendingTasks(self):
        """
        Check if taks are present in pending queue

        :param self: self reference

        :warning: results may be misleading if elements put into the queue are big

        """
        return not self.__pendingQueue.empty()

    def isFull(self):
        """
        Check in peding queue is full

        :param self: self reference

        :warning: results may be misleading if elements put into the queue are big

        """
        return self.__pendingQueue.full()

    def isWorking(self):
        """
        Check existence of working subprocesses

        :param self: self reference
        """
        return not self.__pendingQueue.empty() or self.getNumWorkingProcesses()

    def processResults(self):
        """
        Execute tasks' callbacks removing them from results queue

        :param self: self reference
        """
        processed = 0
        log = sLog.getSubLogger("WorkingProcess")
        while True:
            if (
                not log.debug(
                    "Start loop (t=0) queue size = %d, processed = %d" % (self.__resultsQueue.qsize(), processed)
                )
                and processed == 0
                and self.__resultsQueue.qsize()
            ):
                log.debug("Process results, queue size = %d" % self.__resultsQueue.qsize())
            start = time.time()
            self.__cleanDeadProcesses()
            log.debug("__cleanDeadProcesses", "t=%.2f" % (time.time() - start))
            if not self.__pendingQueue.empty():
                self.__spawnNeededWorkingProcesses()
                log.debug("__spawnNeededWorkingProcesses", "t=%.2f" % (time.time() - start))
            time.sleep(0.1)
            if self.__resultsQueue.empty():
                if self.__resultsQueue.qsize():
                    log.warn("Results queue is empty but has non zero size", "%d" % self.__resultsQueue.qsize())
                    # We only commit suicide if we reach a backlog greater than the maximum number of workers
                    if self.__resultsQueue.qsize() > self.__maxSize:
                        return -1
                    else:
                        return 0
                if processed == 0:
                    log.debug("Process results, but queue is empty...")
                break
            # get task
            task = self.__resultsQueue.get()
            log.debug("__resultsQueue.get", "t=%.2f" % (time.time() - start))
            # execute callbacks
            try:
                task.doExceptionCallback()
                task.doCallback()
                log.debug("doCallback", "t=%.2f" % (time.time() - start))
                if task.usePoolCallbacks():
                    if self.__poolExceptionCallback and task.exceptionRaised():
                        self.__poolExceptionCallback(task.getTaskID(), task.taskException())
                    if self.__poolCallback and task.taskResults():
                        self.__poolCallback(task.getTaskID(), task.taskResults())
                        log.debug("__poolCallback", "t=%.2f" % (time.time() - start))
            except Exception as error:
                log.exception("Exception in callback", lException=error)
                pass
            processed += 1
        if processed:
            log.debug("Processed %d results" % processed)
        else:
            log.debug("No results processed")
        return processed

    def processAllResults(self, timeout=10):
        """
        Process all enqueued tasks at once

        :param self: self reference
        """
        start = time.time()
        while self.getNumWorkingProcesses() or not self.__pendingQueue.empty():
            self.processResults()
            time.sleep(1)
            if time.time() - start > timeout:
                break
        self.processResults()

    def finalize(self, timeout=60):
        """
        Drain pool, shutdown processing in more or less clean way

        :param self: self reference
        :param timeout: seconds to wait before killing
        """
        # start drainig
        self.__draining = True
        # join deamon process
        if self.__daemonProcess:
            self.__daemonProcess.join(timeout)
        # process all tasks
        self.processAllResults(timeout)
        # set stop event, all idle workers should be terminated
        self.__stopEvent.set()
        # join idle workers
        start = time.time()
        log = sLog.getSubLogger("finalize")
        nWorkers = 9999999
        while self.__workersDict:
            self.__cleanDeadProcesses()
            if len(self.__workersDict) != nWorkers:
                nWorkers = len(self.__workersDict)
                log.debug("%d workers still active, timeout = %d" % (nWorkers, timeout))
            if timeout <= 0 or time.time() - start >= timeout:
                break
            time.sleep(0.1)
        # second clean up - join and terminate workers
        if self.__workersDict:
            log.debug(
                "After cleaning dead processes, %d workers still active, timeout = %d"
                % (len(self.__workersDict), timeout)
            )
        for worker in self.__workersDict.values():
            if worker.is_alive():
                worker.terminate()
                worker.join(5)
        self.__cleanDeadProcesses()
        # third clean up - kill'em all!!!
        if self.__workersDict:
            log.debug(
                "After terminating processes, %d workers still active, timeout = %d, kill them"
                % (len(self.__workersDict), timeout)
            )
        self.__filicide()

    def __filicide(self):
        """
        Kill all workers, kill'em all!

        :param self: self reference
        """
        while self.__workersDict:
            pid = list(self.__workersDict).pop(0)
            worker = self.__workersDict[pid]
            if worker.is_alive():
                os.kill(pid, signal.SIGKILL)
            del self.__workersDict[pid]

    def daemonize(self):
        """
        Make ProcessPool a finite being for opening and closing doors between
        chambers.
        Also just run it in a separate background thread to the death of
        PID 0.

        :param self: self reference
        """
        if self.__daemonProcess:
            return
        self.__daemonProcess = threading.Thread(target=self.__backgroundProcess)
        self.__daemonProcess.setDaemon(1)
        self.__daemonProcess.start()

    def __backgroundProcess(self):
        """
        Daemon thread target

        :param self: self reference
        """
        while True:
            if self.__draining:
                return
            self.processResults()
            time.sleep(1)

    def __del__(self):
        """
        Delete slot

        :param self: self reference
        """
        self.finalize(timeout=10)



""" CStoJSONSynchronizer
  Module that keeps the pilot parameters file synchronized with the information
  in the Operations/Pilot section of the CS. If there are additions in the CS,
  these are incorporated to the file.
  The module uploads to a web server the latest version of the pilot scripts.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


__RCSID__ = "$Id$"

import os
import glob
import shutil
import tarfile
import datetime

from six.moves.urllib.request import urlopen

from git import Repo

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath


class PilotCStoJSONSynchronizer(object):
    """
    2 functions are executed:
    - It updates a JSON file with the values on the CS which can be used by Pilot3 pilots
    - It updates the pilot 3 files
    This synchronizer can be triggered at any time via PilotCStoJSONSynchronizer().sync().
    As it is today, this is triggered every time there is a successful write on the CS.
    """

    def __init__(self):
        """c'tor
        Just setting defaults
        """
        self.workDir = ""  # Working directory where the files are going to be stored

        # domain name of the web server(s) used to upload the pilot json file and the pilot scripts
        self.pilotFileServer = ""

        # pilot sync default parameters
        self.pilotRepo = "https://github.com/DIRACGrid/Pilot.git"  # repository of the pilot
        self.pilotVORepo = ""  # repository of the VO that can contain a pilot extension
        self.pilotSetup = gConfig.getValue("/DIRAC/Setup", "")
        self.projectDir = ""
        # where the find the pilot scripts in the VO pilot repository
        self.pilotScriptPath = "Pilot"  # where the find the pilot scripts in the pilot repository
        self.pilotVOScriptPath = ""
        self.pilotRepoBranch = "master"
        self.pilotVORepoBranch = "master"

        self.log = gLogger.getSubLogger(__name__)

        ops = Operations()

        # Overriding parameters from the CS
        self.pilotRepo = ops.getValue("Pilot/pilotRepo", self.pilotRepo)
        self.pilotVORepo = ops.getValue("Pilot/pilotVORepo", self.pilotVORepo)
        self.projectDir = ops.getValue("Pilot/projectDir", self.projectDir)
        self.pilotScriptPath = ops.getValue("Pilot/pilotScriptsPath", self.pilotScriptPath)
        self.pilotVOScriptPath = ops.getValue("Pilot/pilotVOScriptsPath", self.pilotVOScriptPath)
        self.pilotRepoBranch = ops.getValue("Pilot/pilotRepoBranch", self.pilotRepoBranch)
        self.pilotVORepoBranch = ops.getValue("Pilot/pilotVORepoBranch", self.pilotVORepoBranch)

    def getCSDict(self, includeMasterCS=True):
        """Gets minimal info for running a pilot, from the CS

        :returns: pilotDict (containing pilots run info)
        :rtype: S_OK, S_ERROR, value is pilotDict
        """

        pilotDict = {
            "timestamp": datetime.datetime.utcnow().isoformat(),
            "Setups": {},
            "CEs": {},
            "GenericPilotDNs": [],
        }

        self.log.info("-- Getting the content of the CS --")

        # These are in fact not only setups: they may be "Defaults" sections, or VOs, in multi-VOs installations
        setupsRes = gConfig.getSections("/Operations/")
        if not setupsRes["OK"]:
            self.log.error("Can't get sections from Operations", setupsRes["Message"])
            return setupsRes
        setupsInOperations = setupsRes["Value"]

        # getting the setup(s) in this CS, and comparing with what we found in Operations
        setupsInDIRACRes = gConfig.getSections("DIRAC/Setups")
        if not setupsInDIRACRes["OK"]:
            self.log.error("Can't get sections from DIRAC/Setups", setupsInDIRACRes["Message"])
            return setupsInDIRACRes
        setupsInDIRAC = setupsInDIRACRes["Value"]

        # Handling the case of multi-VO CS
        if not set(setupsInDIRAC).intersection(set(setupsInOperations)):
            vos = list(setupsInOperations)
            for vo in vos:
                setupsFromVOs = gConfig.getSections("/Operations/%s" % vo)
                if not setupsFromVOs["OK"]:
                    continue
                else:
                    setupsInOperations = setupsFromVOs["Value"]

        self.log.verbose("From Operations/[Setup]/Pilot")

        for setup in setupsInOperations:
            self._getPilotOptionsPerSetup(setup, pilotDict)

        self.log.verbose("From Resources/Sites")
        sitesSection = gConfig.getSections("/Resources/Sites/")
        if not sitesSection["OK"]:
            self.log.error("Can't get sections from Resources", sitesSection["Message"])
            return sitesSection

        for grid in sitesSection["Value"]:
            gridSection = gConfig.getSections("/Resources/Sites/" + grid)
            if not gridSection["OK"]:
                self.log.error("Can't get sections from Resources", gridSection["Message"])
                return gridSection

            for site in gridSection["Value"]:
                ceList = gConfig.getSections(cfgPath("/Resources", "Sites", grid, site, "CEs"))
                if not ceList["OK"]:
                    # Skip but log it
                    self.log.error("Site has no CEs! - skipping", site)
                    continue

                for ce in ceList["Value"]:
                    # This CEType is like 'HTCondor' or 'ARC' etc.
                    ceType = gConfig.getValue(cfgPath("/Resources", "Sites", grid, site, "CEs", ce, "CEType"))
                    if ceType is None:
                        # Skip but log it
                        self.log.error("CE has no option CEType!", ce + " at " + site)
                        pilotDict["CEs"][ce] = {"Site": site}
                    else:
                        pilotDict["CEs"][ce] = {"Site": site, "GridCEType": ceType}

                    # This LocalCEType is like 'InProcess' or 'Pool' or 'Pool/Singularity' etc.
                    # It can be in the queue and/or the CE level
                    localCEType = gConfig.getValue(cfgPath("/Resources", "Sites", grid, site, "CEs", ce, "LocalCEType"))
                    if localCEType is not None:
                        pilotDict["CEs"][ce].setdefault("LocalCEType", localCEType)

                    res = gConfig.getSections(cfgPath("/Resources", "Sites", grid, site, "CEs", ce, "Queues"))
                    if not res["OK"]:
                        # Skip but log it
                        self.log.error("No queues found for CE", ce + ": " + res["Message"])
                        continue
                    queueList = res["Value"]
                    for queue in queueList:
                        localCEType = gConfig.getValue(
                            cfgPath("/Resources", "Sites", grid, site, "CEs", ce, "Queues", queue, "LocalCEType")
                        )
                        if localCEType is not None:
                            pilotDict["CEs"][ce].setdefault(queue, {"LocalCEType": localCEType})

        defaultSetup = gConfig.getValue("/DIRAC/DefaultSetup")
        if defaultSetup:
            pilotDict["DefaultSetup"] = defaultSetup

        self.log.debug("From DIRAC/Configuration")
        configurationServers = gConfig.getServersList()
        if not includeMasterCS:
            masterCS = gConfigurationData.getMasterServer()
            configurationServers = list(set(configurationServers) - set([masterCS]))
        pilotDict["ConfigurationServers"] = configurationServers

        self.log.debug("Got pilotDict", str(pilotDict))

        return S_OK(pilotDict)

    def _getPilotOptionsPerSetup(self, setup, pilotDict):
        """Given a setup, returns its pilot options in a dictionary"""

        options = gConfig.getOptionsDict("/Operations/%s/Pilot" % setup)
        if not options["OK"]:
            self.log.warn("Section does not exist: skipping", "/Operations/%s/Pilot " % setup)
            return

        # We include everything that's in the Pilot section for this setup
        if setup == self.pilotSetup:
            self.pilotVOVersion = options["Value"]["Version"]
        pilotDict["Setups"][setup] = options["Value"]
        # We update separately 'GenericPilotDNs'
        try:
            pilotDict["GenericPilotDNs"].append(pilotDict["Setups"][setup]["GenericPilotDN"])
        except KeyError:
            pass
        ceTypesCommands = gConfig.getOptionsDict("/Operations/%s/Pilot/Commands" % setup)
        if ceTypesCommands["OK"]:
            # It's ok if the Pilot section doesn't list any Commands too
            pilotDict["Setups"][setup]["Commands"] = {}
            for ceType in ceTypesCommands["Value"]:
                # FIXME: inconsistent that we break Commands down into a proper list but other things are comma-list strings
                pilotDict["Setups"][setup]["Commands"][ceType] = ceTypesCommands["Value"][ceType].split(", ")
                # pilotDict['Setups'][setup]['Commands'][ceType] = ceTypesCommands['Value'][ceType]
        if "CommandExtensions" in pilotDict["Setups"][setup]:
            # FIXME: inconsistent that we break CommandExtensionss down into a proper
            # list but other things are comma-list strings
            pilotDict["Setups"][setup]["CommandExtensions"] = pilotDict["Setups"][setup]["CommandExtensions"].split(
                ", "
            )
            # pilotDict['Setups'][setup]['CommandExtensions'] = pilotDict['Setups'][setup]['CommandExtensions']

        # Getting the details aboout the MQ Services to be used for logging, if any
        if "LoggingMQService" in pilotDict["Setups"][setup]:
            loggingMQService = gConfig.getOptionsDict(
                "/Resources/MQServices/%s" % pilotDict["Setups"][setup]["LoggingMQService"]
            )
            if not loggingMQService["OK"]:
                self.log.error(loggingMQService["Message"])
                return loggingMQService
            pilotDict["Setups"][setup]["Logging"] = {}
            pilotDict["Setups"][setup]["Logging"]["Host"] = loggingMQService["Value"]["Host"]
            pilotDict["Setups"][setup]["Logging"]["Port"] = loggingMQService["Value"]["Port"]

            loggingMQServiceQueuesSections = gConfig.getSections(
                "/Resources/MQServices/%s/Queues" % pilotDict["Setups"][setup]["LoggingMQService"]
            )
            if not loggingMQServiceQueuesSections["OK"]:
                self.log.error(loggingMQServiceQueuesSections["Message"])
                return loggingMQServiceQueuesSections
            pilotDict["Setups"][setup]["Logging"]["Queue"] = {}

            for queue in loggingMQServiceQueuesSections["Value"]:
                loggingMQServiceQueue = gConfig.getOptionsDict(
                    "/Resources/MQServices/%s/Queues/%s" % (pilotDict["Setups"][setup]["LoggingMQService"], queue)
                )
                if not loggingMQServiceQueue["OK"]:
                    self.log.error(loggingMQServiceQueue["Message"])
                    return loggingMQServiceQueue
                pilotDict["Setups"][setup]["Logging"]["Queue"][queue] = loggingMQServiceQueue["Value"]

            queuesRes = gConfig.getSections(
                "/Resources/MQServices/%s/Queues" % pilotDict["Setups"][setup]["LoggingMQService"]
            )
            if not queuesRes["OK"]:
                return queuesRes
            queues = queuesRes["Value"]
            queuesDict = {}
            for queue in queues:
                queueOptionRes = gConfig.getOptionsDict(
                    "/Resources/MQServices/%s/Queues/%s" % (pilotDict["Setups"][setup]["LoggingMQService"], queue)
                )
                if not queueOptionRes["OK"]:
                    return queueOptionRes
                queuesDict[queue] = queueOptionRes["Value"]
            pilotDict["Setups"][setup]["Logging"]["Queues"] = queuesDict

    def syncScripts(self):
        """Clone the pilot scripts from the Pilot repositories (handle also extensions)"""
        tarFiles = []

        # Extension, if it exists
        if self.pilotVORepo:
            pilotVOLocalRepo = os.path.join(self.workDir, "pilotVOLocalRepo")
            if os.path.isdir(pilotVOLocalRepo):
                shutil.rmtree(pilotVOLocalRepo)
            os.mkdir(pilotVOLocalRepo)
            repo_VO = Repo.init(pilotVOLocalRepo)
            upstream = repo_VO.create_remote("upstream", self.pilotVORepo)
            upstream.fetch()
            upstream.pull(upstream.refs[0].remote_head)
            if repo_VO.tags:
                repo_VO.git.checkout(repo_VO.tags[self.pilotVOVersion], b="pilotVOScripts")
            else:
                repo_VO.git.checkout("upstream/%s" % self.pilotVORepoBranch, b="pilotVOScripts")
            scriptDir = os.path.join(pilotVOLocalRepo, self.projectDir, self.pilotVOScriptPath, "*.py")
            for fileVO in glob.glob(scriptDir):
                tarFiles.append(fileVO)
        else:
            self.log.info("The /Operations/<Setup>/Pilot/pilotVORepo option is not defined, using Vanilla DIRAC pilot")

        # DIRAC repo
        pilotLocalRepo = os.path.join(self.workDir, "pilotLocalRepo")
        if os.path.isdir(pilotLocalRepo):
            shutil.rmtree(pilotLocalRepo)
        os.mkdir(pilotLocalRepo)
        repo = Repo.init(pilotLocalRepo)
        upstream = repo.create_remote("upstream", self.pilotRepo)
        upstream.fetch()
        upstream.pull(upstream.refs[0].remote_head)
        repo.git.checkout("upstream/%s" % self.pilotRepoBranch, b="pilotScripts")

        scriptDir = os.path.join(pilotLocalRepo, self.pilotScriptPath, "*.py")
        for filename in glob.glob(scriptDir):
            tarFiles.append(filename)
        if not os.path.isfile(os.path.join(pilotLocalRepo, self.pilotScriptPath, "dirac-install.py")):

            # Download dirac-install.py
            response = urlopen("https://raw.githubusercontent.com/DIRACGrid/management/master/dirac-install.py")
            code = response.getcode()
            if code > 200 or code >= 300:
                return S_ERROR("Failed to download dirac-install.py with code %s" % code)
            with open("dirac-install.py", "wb") as fp:
                fp.write(response.read())

            tarFiles.append("dirac-install.py")

        tarPath = os.path.join(self.workDir, "pilot.tar")
        with tarfile.TarFile(name=tarPath, mode="w") as tf:
            for ptf in tarFiles:
                # This copy makes sure that all the files in the tarball are accessible
                # in the work directory. It should be kept
                shutil.copyfile(ptf, os.path.join(self.workDir, os.path.basename(ptf)))
                tf.add(ptf, arcname=os.path.basename(ptf), recursive=False)

        tarFilesPaths = [os.path.join(self.workDir, os.path.basename(tarredF)) for tarredF in tarFiles]

        return S_OK((tarPath, tarFilesPaths))



"""
Test backend attachment
"""

__RCSID__ = "$Id$"

import pytest

from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import gLogger, gLoggerReset, cleaningLog


def getContentFromFilename(backendOptions):
    """Get content from the file attached to a given backend and erase the content from the file."""
    filename = backendOptions.get("FileName")
    if not filename:
        return None

    # get the content of the file
    with open(filename, "r") as fileContent:
        content = fileContent.read()

    # clean the content
    lines = content.split("\n")
    cleanContent = ""
    for line in lines:
        cleanContent += cleaningLog(line)

    # reset the file
    with open(filename, "w") as fileContent:
        pass

    return cleanContent


@pytest.mark.parametrize(
    "backends",
    [
        (
            {
                "file1": {
                    "logger": "gLogger",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "Framework NOTICE: msgFramework/log NOTICE: msgFramework/log/sublog NOTICE: msg",
                }
            }
        ),
        (
            {
                "file2": {
                    "logger": "log",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "Framework/log NOTICE: msgFramework/log/sublog NOTICE: msg",
                }
            }
        ),
        (
            {
                "file3": {
                    "logger": "sublog",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "Framework/log/sublog NOTICE: msg",
                }
            }
        ),
        (
            {
                "file4": {
                    "logger": "gLogger",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp", "LogLevel": "error"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "",
                }
            }
        ),
        (
            {
                "file5": {
                    "logger": "log",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp", "LogLevel": "error"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "",
                }
            }
        ),
        (
            {
                "file6": {
                    "logger": "sublog",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp", "LogLevel": "error"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "",
                }
            }
        ),
        (
            {
                "file7a": {
                    "logger": "gLogger",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test1.tmp"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "Framework NOTICE: msgFramework/log NOTICE: msgFramework/log/sublog NOTICE: msg",
                },
                "file7b": {
                    "logger": "gLogger",
                    "backendType": "file",
                    "backendOptions": {"FileName": "backend_test2.tmp"},
                    "extractBackendContent": getContentFromFilename,
                    "backendContent": "Framework NOTICE: msgFramework/log NOTICE: msgFramework/log/sublog NOTICE: msg",
                },
            }
        ),
    ],
)
def test_registerBackendgLogger(backends):
    """
    Attach backends to gLogger, generate some logs from different loggers and check the content of the backends
    """
    _, log, sublog = gLoggerReset()

    # dictionary of available loggers
    loggers = {"gLogger": gLogger, "log": log, "sublog": sublog}

    # attach backends to the corresponding logger
    for backend, params in backends.items():
        logger = loggers[params["logger"]]
        numberOfBackends = len(logger._backendsList)
        logger.registerBackend(params["backendType"], params["backendOptions"])

        # backend should be added to logger.backendList
        assert len(logger._backendsList) == (numberOfBackends + 1)

    # Generate logs from gLogger, log, sublog
    gLogger.setLevel("notice")
    gLogger.notice("msg")
    log.notice("msg")
    sublog.notice("msg")

    # Check the content of the backends
    for backend, params in backends.items():
        content = params["extractBackendContent"](params["backendOptions"])
        assert content == params["backendContent"]



#! /usr/bin/env python
########################################################################
# File :    dirac-wms-jobs-select-output-search
# Author :  Vladimir Romanovsky
########################################################################
"""
Retrieve output sandbox for DIRAC Jobs for the given selection and search for a string in their std.out
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
from shutil import rmtree

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("", "Status=", "Primary status")
    Script.registerSwitch("", "MinorStatus=", "Secondary status")
    Script.registerSwitch("", "ApplicationStatus=", "Application status")
    Script.registerSwitch("", "Site=", "Execution site")
    Script.registerSwitch("", "Owner=", "Owner (DIRAC nickname)")
    Script.registerSwitch("", "JobGroup=", "Select jobs for specified job group")
    Script.registerSwitch("", "Date=", "Date in YYYY-MM-DD format, if not specified default is today")
    Script.registerSwitch("", "File=", "File name,if not specified default is std.out ")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("String: string to search for")
    _, args = Script.parseCommandLine(ignoreErrors=True)

    # Default values
    status = None
    minorStatus = None
    appStatus = None
    site = None
    owner = None
    jobGroup = None
    date = None
    filename = "std.out"

    if len(args) != 1:
        Script.showHelp()

    searchstring = str(args[0])

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "status":
            status = switch[1]
        elif switch[0].lower() == "minorstatus":
            minorStatus = switch[1]
        elif switch[0].lower() == "applicationstatus":
            appStatus = switch[1]
        elif switch[0].lower() == "site":
            site = switch[1]
        elif switch[0].lower() == "owner":
            owner = switch[1]
        elif switch[0].lower() == "jobgroup":
            jobGroup = switch[1]
        elif switch[0].lower() == "date":
            date = switch[1]
        elif switch[0].lower() == "file":
            filename = switch[1]

    selDate = date
    if not date:
        selDate = "Today"

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0
    errorList = []
    resultDict = {}

    result = dirac.selectJobs(
        status=status,
        minorStatus=minorStatus,
        applicationStatus=appStatus,
        site=site,
        owner=owner,
        jobGroup=jobGroup,
        date=date,
    )
    if result["OK"]:
        jobs = result["Value"]
    else:
        print("Error in selectJob", result["Message"])
        DIRAC.exit(2)

    for job in jobs:

        result = dirac.getOutputSandbox(job)
        if result["OK"]:
            if os.path.exists("%s" % job):

                lines = []
                try:
                    lines = open(os.path.join(job, filename)).readlines()
                except Exception as x:
                    errorList.append((job, x))
                for line in lines:
                    if line.count(searchstring):
                        resultDict[job] = line
                rmtree("%s" % (job))
        else:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for result in resultDict.items():
        print(result)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" DIRAC FileCatalog Database """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import errno

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Resources.Catalog.Utilities import checkArgumentFormat
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

#############################################################################


class FileCatalogDB(DB):
    def __init__(self, databaseLocation="DataManagement/FileCatalogDB"):
        # The database location can be specified in System/Database form or in just the Database name
        # in the DataManagement system
        db = databaseLocation
        if "/" not in db:
            db = "DataManagement/" + db

        super(FileCatalogDB, self).__init__("FileCatalogDB", db)

        self.ugManager = None
        self.seManager = None
        self.securityManager = None
        self.dtree = None
        self.fileManager = None
        self.dmeta = None
        self.fmeta = None
        self.datasetManager = None

    def setConfig(self, databaseConfig):
        self.directories = {}
        # In memory storage of the various parameters
        self.users = {}
        self.uids = {}
        self.groups = {}
        self.gids = {}
        self.seNames = {}
        self.seids = {}

        # Obtain some general configuration of the database
        self.uniqueGUID = databaseConfig["UniqueGUID"]
        self.globalReadAccess = databaseConfig["GlobalReadAccess"]
        self.lfnPfnConvention = databaseConfig["LFNPFNConvention"]
        if self.lfnPfnConvention == "None":
            self.lfnPfnConvention = False
        self.resolvePfn = databaseConfig["ResolvePFN"]
        self.umask = databaseConfig["DefaultUmask"]
        self.validFileStatus = databaseConfig["ValidFileStatus"]
        self.validReplicaStatus = databaseConfig["ValidReplicaStatus"]
        self.visibleFileStatus = databaseConfig["VisibleFileStatus"]
        self.visibleReplicaStatus = databaseConfig["VisibleReplicaStatus"]

        # Load the configured components
        for compAttribute, componentType in [
            ("ugManager", "UserGroupManager"),
            ("seManager", "SEManager"),
            ("securityManager", "SecurityManager"),
            ("dtree", "DirectoryManager"),
            ("fileManager", "FileManager"),
            ("datasetManager", "DatasetManager"),
            ("dmeta", "DirectoryMetadata"),
            ("fmeta", "FileMetadata"),
        ]:

            result = self.__loadCatalogComponent(componentType, databaseConfig[componentType])
            if not result["OK"]:
                return result
            self.__setattr__(compAttribute, result["Value"])

        return S_OK()

    def __loadCatalogComponent(self, componentType, componentName):
        """Create an object of a given catalog component"""
        componentModule = "DataManagementSystem.DB.FileCatalogComponents.%s.%s" % (componentType, componentName)
        result = ObjectLoader().loadObject(componentModule)
        if not result["OK"]:
            gLogger.error("Failed to load catalog component", "%s: %s" % (componentName, result["Message"]))
            return result
        componentClass = result["Value"]
        component = componentClass(self)
        return S_OK(component)

    def setUmask(self, umask):
        self.umask = umask

    ########################################################################
    #
    #  SE based write methods
    #

    def addSE(self, seName, credDict):
        """
        Add a new StorageElement

        :param str seName: Name of the StorageElement
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.seManager.addSE(seName)

    def deleteSE(self, seName, credDict):
        """
        Delete a StorageElement

        :param str seName: Name of the StorageElement
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.seManager.deleteSE(seName)

    ########################################################################
    #
    #  User/groups based write methods
    #

    def addUser(self, userName, credDict):
        """
        Add a new user

        :param str userName: Name of the User
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.addUser(userName)

    def deleteUser(self, userName, credDict):
        """
        Delete a user

        :param str userName: Name of the User
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.deleteUser(userName)

    def addGroup(self, groupName, credDict):
        """
        Add a new group

        :param str groupName: Name of the group
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.addGroup(groupName)

    def deleteGroup(self, groupName, credDict):
        """
        Delete a group

        :param str groupName: Name of the group
        :param credDict: credential
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.deleteGroup(groupName)

    ########################################################################
    #
    #  User/groups based read methods
    #

    def getUsers(self, credDict):
        """
        Returns the list of users

        :param credDict: credential
        :return: dictionary indexed on the user name
        """
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.getUsers()

    def getGroups(self, credDict):
        """
        Returns the list of groups

        :param credDict: credential
        :return: dictionary indexed on the group name
        """

        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        return self.ugManager.getGroups()

    ########################################################################
    #
    #  Path based read methods
    #

    def exists(self, lfns, credDict):
        res = self._checkPathPermissions("exists", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        if res["Value"]["Successful"]:
            res = self.fileManager.exists(res["Value"]["Successful"])
            if not res["OK"]:
                return res
            failed.update(res["Value"]["Failed"])
            successful = res["Value"]["Successful"]

            notExist = []
            for lfn in list(res["Value"]["Successful"]):
                if not successful[lfn]:
                    notExist.append(lfn)
                    successful.pop(lfn)
            if notExist:
                res = self.dtree.exists(notExist)
                if not res["OK"]:
                    return res
                failed.update(res["Value"]["Failed"])
                successful.update(res["Value"]["Successful"])

        return S_OK({"Successful": successful, "Failed": failed})

    def getPathPermissions(self, lfns, credDict):
        """Get permissions for the given user/group to manipulate the given lfns"""
        res = checkArgumentFormat(lfns)
        if not res["OK"]:
            return res
        lfns = res["Value"]

        return self.securityManager.getPathPermissions(list(lfns), credDict)

    def hasAccess(self, opType, paths, credDict):
        """Get permissions for the given user/group to execute the given operation
        on the given paths

        returns Successful dict with True/False
        """
        res = checkArgumentFormat(paths)
        if not res["OK"]:
            return res
        paths = res["Value"]

        return self.securityManager.hasAccess(opType, paths, credDict)

    ########################################################################
    #
    #  Path based read methods
    #

    def changePathOwner(self, paths, credDict, recursive=False):
        """Bulk method to change Owner for the given paths

        :param dict paths: dictionary < lfn : owner >
        :param dict credDict: dictionary of the caller credentials
        :param boolean recursive: flag to apply the operation recursively
        """
        res = self._checkPathPermissions("changePathOwner", paths, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        if res["Value"]["Successful"]:
            result = self.__changePathFunction(
                res["Value"]["Successful"],
                credDict,
                self.dtree.changeDirectoryOwner,
                self.fileManager.changeFileOwner,
                recursive=recursive,
            )
            failed.update(result["Value"]["Failed"])
            successful = result["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def changePathGroup(self, paths, credDict, recursive=False):
        """Bulk method to change Group for the given paths

        :param dict paths: dictionary < lfn : group >
        :param dict credDict: dictionary of the caller credentials
        :param boolean recursive: flag to apply the operation recursively
        """
        res = self._checkPathPermissions("changePathGroup", paths, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        if res["Value"]["Successful"]:
            result = self.__changePathFunction(
                res["Value"]["Successful"],
                credDict,
                self.dtree.changeDirectoryGroup,
                self.fileManager.changeFileGroup,
                recursive=recursive,
            )
            failed.update(result["Value"]["Failed"])
            successful = result["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def changePathMode(self, paths, credDict, recursive=False):
        """Bulk method to change Mode for the given paths

        :param dict paths: dictionary < lfn : mode >
        :param dict credDict: dictionary of the caller credentials
        :param boolean recursive: flag to apply the operation recursively
        """

        res = self._checkPathPermissions("changePathMode", paths, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        if res["Value"]["Successful"]:
            result = self.__changePathFunction(
                res["Value"]["Successful"],
                credDict,
                self.dtree.changeDirectoryMode,
                self.fileManager.changeFileMode,
                recursive=recursive,
            )
            failed.update(result["Value"]["Failed"])
            successful = result["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def __changePathFunction(self, paths, credDict, change_function_directory, change_function_file, recursive=False):
        """A generic function to change Owner, Group or Mode for the given paths

        :param dict paths: dictionary < lfn : parameter_value >
        :param dict credDict: dictionary of the caller credentials
        :param function change_function_directory: function to change directory parameters
        :param function change_function_file: function to change file parameters
        :param boolean recursive: flag to apply the operation recursively
        """
        dirList = []
        result = self.isDirectory(paths, credDict)
        if not result["OK"]:
            return result
        for di in result["Value"]["Successful"]:
            if result["Value"]["Successful"][di]:
                dirList.append(di)
        fileList = []
        if len(dirList) < len(paths):
            result = self.isFile(paths, credDict)
            if not result["OK"]:
                return result
            for fi in result["Value"]["Successful"]:
                if result["Value"]["Successful"][fi]:
                    fileList.append(fi)

        successful = {}
        failed = {}

        dirArgs = {}
        fileArgs = {}

        for path in paths:
            if (path not in dirList) and (path not in fileList):
                failed[path] = "No such file or directory"
            if path in dirList:
                dirArgs[path] = paths[path]
            elif path in fileList:
                fileArgs[path] = paths[path]
        if dirArgs:
            result = change_function_directory(dirArgs, recursive=recursive)
            if not result["OK"]:
                return result
            successful.update(result["Value"]["Successful"])
            failed.update(result["Value"]["Failed"])
        if fileArgs:
            result = change_function_file(fileArgs)
            if not result["OK"]:
                return result
            successful.update(result["Value"]["Successful"])
            failed.update(result["Value"]["Failed"])
        return S_OK({"Successful": successful, "Failed": failed})

    ########################################################################
    #
    #  File based write methods
    #

    def addFile(self, lfns, credDict):
        """
        Add a new File

        :param dict lfns: indexed on file's LFN, the values are dictionaries which contains
                          the attributes of the files (PFN, SE, Size, GUID, Checksum)
        :param creDict: credential

        :return: Successful/Failed dict.
        """
        res = self._checkPathPermissions("addFile", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.addFile(res["Value"]["Successful"], credDict)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def setFileStatus(self, lfns, credDict):
        """
        Set the status of a File

        :param dict lfns: dict indexed on the LFNs. The values are the status (should be in config['ValidFileStatus'])
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("setFileStatus", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.setFileStatus(res["Value"]["Successful"], credDict)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def removeFile(self, lfns, credDict):
        """
         Remove files

        :param lfns: list of LFNs to remove
        :type lfns: python:list
        :param creDict: credential
        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("removeFile", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.removeFile(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def addReplica(self, lfns, credDict):
        """
         Add a replica to a File

        :param dict lfns: keys are LFN. The values are dict with key PFN and SE
                          (e.g. {myLfn : {"PFN" : "myPfn", "SE" : "mySE"}})
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("addReplica", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.addReplica(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def removeReplica(self, lfns, credDict):
        """
         Remove replicas

        :param dict lfns: keys are LFN. The values are dict with key PFN and SE
                          (e.g. {myLfn : {"PFN" : "myPfn", "SE" : "mySE"}})
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("removeReplica", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.removeReplica(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def setReplicaStatus(self, lfns, credDict):
        """
        Set the status of a Replicas

        :param dict lfns: dict indexed on the LFNs. The values are dict with keys
                          "SE" and "Status" (that has to be in config['ValidReplicaStatus'])
        :param creDict: credential

        :return: Successful/Failed dict.
        """
        res = self._checkPathPermissions("setReplicaStatus", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.setReplicaStatus(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def setReplicaHost(self, lfns, credDict):
        res = self._checkPathPermissions("setReplicaHost", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.setReplicaHost(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def addFileAncestors(self, lfns, credDict):
        """Add ancestor information for the given LFNs"""
        res = self._checkPathPermissions("addFileAncestors", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.addFileAncestors(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    ########################################################################
    #
    #  File based read methods
    #

    def isFile(self, lfns, credDict):
        """
        Checks whether a list of LFNS are files or not

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
                The values of the successful dict are True or False whether it's a file or not
        """

        res = self._checkPathPermissions("isFile", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.isFile(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getFileSize(self, lfns, credDict):
        """
        Gets the size of a list of lfns

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("getFileSize", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getFileSize(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getFileMetadata(self, lfns, credDict):
        """
        Gets the metadata of a list of lfns

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("getFileMetadata", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getFileMetadata(res["Value"]["Successful"])

        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getReplicas(self, lfns, allStatus, credDict):
        """
        Gets the list of replicas of a list of lfns

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param allStatus: if all the status are visible, or only those defined in config['ValidReplicaStatus']
        :param creDict: credential

        :return: Successful/Failed dict. Successful is indexed on the LFN, and the values are dictionary
                 with the SEName as keys
        """

        res = self._checkPathPermissions("getReplicas", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getReplicas(res["Value"]["Successful"], allStatus=allStatus)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getReplicaStatus(self, lfns, credDict):
        """
        Gets the status of a list of replicas

        :param dict lfns: <lfn, se name>
        :param creDict: credential

        :return: Successful/Failed dict.
        """

        res = self._checkPathPermissions("getReplicaStatus", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getReplicaStatus(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getFileAncestors(self, lfns, depths, credDict):
        res = self._checkPathPermissions("getFileAncestors", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getFileAncestors(res["Value"]["Successful"], depths)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getFileDescendents(self, lfns, depths, credDict):
        res = self._checkPathPermissions("getFileDescendents", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.fileManager.getFileDescendents(res["Value"]["Successful"], depths)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getFileDetails(self, lfnList, credDict):
        """Get all the metadata for the given files"""
        connection = False
        result = self.fileManager._findFiles(lfnList, connection=connection)
        if not result["OK"]:
            return result
        resultDict = {}
        fileIDDict = {}
        lfnDict = result["Value"]["Successful"]
        for lfn in lfnDict:
            fileIDDict[lfnDict[lfn]["FileID"]] = lfn

        result = self.fileManager._getFileMetadataByID(list(fileIDDict), connection=connection)
        if not result["OK"]:
            return result
        for fileID in result["Value"]:
            resultDict[fileIDDict[fileID]] = result["Value"][fileID]

        result = self.fmeta._getFileUserMetadataByID(list(fileIDDict), credDict, connection=connection)
        if not result["OK"]:
            return result
        for fileID in fileIDDict:
            resultDict[fileIDDict[fileID]].setdefault("Metadata", {})
            if fileID in result["Value"]:
                resultDict[fileIDDict[fileID]]["Metadata"] = result["Value"][fileID]

        return S_OK(resultDict)

    def getLFNForGUID(self, guids, credDict):
        """
        Gets the lfns that match a list of guids

        :param lfns: list of guid to look for
        :type lfns: python:list
        :param creDict: credential

        :return: S_OK({guid:lfn}) dict.
        """
        return self.fileManager.getLFNForGUID(guids)

    ########################################################################
    #
    #  Directory based Write methods
    #

    def createDirectory(self, lfns, credDict):
        """
        Create new directories

        :param lfns: list of directories
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
        """
        res = self._checkPathPermissions("createDirectory", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.createDirectory(res["Value"]["Successful"], credDict)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def removeDirectory(self, lfns, credDict):
        """
        Remove directories

        :param lfns: list of directories
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
        """
        res = self._checkPathPermissions("removeDirectory", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.removeDirectory(res["Value"]["Successful"], credDict)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        if not successful:
            return S_OK({"Successful": successful, "Failed": failed})

        # Remove the directory metadata now
        dirIdList = [successful[p]["DirID"] for p in successful if "DirID" in successful[p]]
        result = self.dmeta.removeMetadataForDirectory(dirIdList, credDict)
        if not result["OK"]:
            return result
        failed.update(result["Value"]["Failed"])
        # We remove from The successful those that failed in the metadata removal
        map(lambda x: successful.pop(x) if x in successful else None, failed)
        # We update the successful
        successful.update(result["Value"]["Successful"])
        return S_OK({"Successful": successful, "Failed": failed})

    ########################################################################
    #
    #  Directory based read methods
    #

    def listDirectory(self, lfns, credDict, verbose=False):
        """
        List directories

        :param lfns: list of directories
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
           The successful values are dictionaries indexed "Files", "Datasets", "Subdirs" and "Links"
        """

        res = self._checkPathPermissions("listDirectory", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.listDirectory(res["Value"]["Successful"], verbose=verbose)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def isDirectory(self, lfns, credDict):
        """
        Checks whether a list of LFNS are directories or not

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param creDict: credential

        :return: Successful/Failed dict.
                The values of the successful dict are True or False whether it's a dir or not
        """

        res = self._checkPathPermissions("isDirectory", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.isDirectory(res["Value"]["Successful"])
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getDirectoryReplicas(self, lfns, allStatus, credDict):
        res = self._checkPathPermissions("getDirectoryReplicas", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.getDirectoryReplicas(res["Value"]["Successful"], allStatus)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def getDirectorySize(self, lfns, longOutput, fromFiles, recursiveSum, credDict):
        """
        Get the sizes of a list of directories

        :param lfns: list of LFN to check
        :type lfns: python:list
        :param longOutput: if True, get also the physical size per SE (takes longer)
        :param fromFiles: if True, recompute the size from the file tables instead of the
                         precomputed values (takes longer)
        :param recursiveSum: if True (default), takes into account the subdirectories
        :param creDict: credential

        :return: Successful/Failed dict.
            The successful values are dictionaries indexed "LogicalFiles" (nb of files),
            "LogicalDirectories" (nb of dir) and "LogicalSize" (sum of File's sizes)
        """

        res = self._checkPathPermissions("getDirectorySize", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]

        # if no successful, just return
        if not res["Value"]["Successful"]:
            return S_OK({"Successful": {}, "Failed": failed})

        res = self.dtree.getDirectorySize(res["Value"]["Successful"], longOutput, fromFiles, recursiveSum)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        successful = res["Value"]["Successful"]
        queryTime = res["Value"].get("QueryTime", -1.0)
        return S_OK({"Successful": successful, "Failed": failed, "QueryTime": queryTime})

    def getDirectoryMetadata(self, lfns, credDict):
        """Get standard directory metadata

        :param lfns: list of directory paths
        :type lfns: python:list
        :param dict credDict: credentials
        :return: Successful/Failed dict.
        """
        res = self._checkPathPermissions("getDirectoryMetadata", lfns, credDict)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn in res["Value"]["Successful"]:
            result = self.dtree.getDirectoryParameters(lfn)
            if result["OK"]:
                successful[lfn] = result["Value"]
            else:
                failed[lfn] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def rebuildDirectoryUsage(self):
        """Rebuild DirectoryUsage table from scratch"""

        result = self.dtree._rebuildDirectoryUsage()
        return result

    def repairCatalog(self, credDict={}):
        """Repair catalog inconsistencies"""

        result = self._checkAdminPermission(credDict)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR(errno.EACCES, "Not authorized to perform catalog repairs")

        resultDict = {}
        resultDict["RecoverOrphanDirectories"] = self.dtree.recoverOrphanDirectories(credDict)
        resultDict["RepairFileTables"] = self.fileManager.repairFileTables()

        return S_OK(resultDict)

    #######################################################################
    #
    #  Catalog metadata methods
    #

    def setMetadata(self, path, metadataDict, credDict):
        """Add metadata to the given path"""
        res = self._checkPathPermissions("setMetadata", path, credDict)
        if not res["OK"]:
            return res
        if not res["Value"]["Successful"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        if not res["Value"]["Successful"][path]:
            return S_ERROR(errno.EACCES, "Permission denied")

        result = self.dtree.isDirectory({path: True})
        if not result["OK"]:
            return result
        if not result["Value"]["Successful"]:
            return S_ERROR("Failed to determine the path type")
        if result["Value"]["Successful"][path]:
            # This is a directory
            return self.dmeta.setMetadata(path, metadataDict, credDict)
        else:
            # This is a file
            return self.fmeta.setMetadata(path, metadataDict, credDict)

    def setMetadataBulk(self, pathMetadataDict, credDict):
        """Add metadata for the given paths"""
        successful = {}
        failed = {}
        for path, metadataDict in pathMetadataDict.items():
            result = self.setMetadata(path, metadataDict, credDict)
            if result["OK"]:
                successful[path] = True
            else:
                failed[path] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def removeMetadata(self, pathMetadataDict, credDict):
        """Remove metadata for the given paths"""
        successful = {}
        failed = {}
        for path, metadataDict in pathMetadataDict.items():
            result = self.__removeMetadata(path, metadataDict, credDict)
            if result["OK"]:
                successful[path] = True
            else:
                failed[path] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    def __removeMetadata(self, path, metadata, credDict):
        """Remove metadata from the given path"""
        res = self._checkPathPermissions("__removeMetadata", path, credDict)
        if not res["OK"]:
            return res
        if not res["Value"]["Successful"]:
            return S_ERROR("Permission denied")
        if not res["Value"]["Successful"][path]:
            return S_ERROR("Permission denied")

        result = self.dtree.isDirectory({path: True})
        if not result["OK"]:
            return result
        if not result["Value"]["Successful"]:
            return S_ERROR("Failed to determine the path type")
        if result["Value"]["Successful"][path]:
            # This is a directory
            return self.dmeta.removeMetadata(path, metadata, credDict)
        else:
            # This is a file
            return self.fmeta.removeMetadata(path, metadata, credDict)

    #######################################################################
    #
    #  Catalog admin methods
    #

    def getCatalogCounters(self, credDict):
        counterDict = {}
        res = self._checkAdminPermission(credDict)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR(errno.EACCES, "Permission denied")
        # res = self.dtree.getDirectoryCounters()
        # if not res['OK']:
        #  return res
        # counterDict.update(res['Value'])
        res = self.fileManager.getFileCounters()
        if not res["OK"]:
            return res
        counterDict.update(res["Value"])
        res = self.fileManager.getReplicaCounters()
        if not res["OK"]:
            return res
        counterDict.update(res["Value"])
        res = self.dtree.getDirectoryCounters()
        if not res["OK"]:
            return res
        counterDict.update(res["Value"])
        return S_OK(counterDict)

    ########################################################################
    #
    #  Security based methods
    #

    def _checkAdminPermission(self, credDict):
        return self.securityManager.hasAdminAccess(credDict)

    def _checkPathPermissions(self, operation, lfns, credDict):

        res = checkArgumentFormat(lfns)
        if not res["OK"]:
            return res
        lfns = res["Value"]

        res = self.securityManager.hasAccess(operation, list(lfns), credDict)
        if not res["OK"]:
            return res
        # Do not consider those paths for which we failed to determine access
        failed = res["Value"]["Failed"]
        for lfn in failed:
            lfns.pop(lfn)
        # Do not consider those paths for which access is denied
        successful = {}
        for lfn, access in res["Value"]["Successful"].items():
            if not access:
                failed[lfn] = "Permission denied"
            else:
                successful[lfn] = lfns[lfn]
        return S_OK({"Successful": successful, "Failed": failed})

    def getSEDump(self, seName):
        """
         Return all the files at a given SE, together with checksum and size

        :param seName: name of the StorageElement

        :returns: S_OK with list of tuples (lfn, checksum, size)
        """
        return self.fileManager.getSEDump(seName)



""" Utilities for the MessageQueue package
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from six.moves import queue as Queue
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI


def getMQParamsFromCS(mqURI):
    """Function gets parameters of a MQ destination (queue/topic) from the CS.

    Args:
      mqURI(str):Pseudo URI identifing the MQ service. It has the following format:
                mqConnection::DestinationType::DestinationName
                e.g. blabla.cern.ch::Queues::MyQueue1
      mType(str): 'consumer' or 'producer'
    Returns:
      S_OK(param_dicts) or S_ERROR
    """
    # API initialization is required to get an up-to-date configuration from the CS
    csAPI = CSAPI()
    csAPI.initialize()

    try:
        mqService, mqType, mqName = mqURI.split("::")
    except ValueError:
        return S_ERROR("Bad format of mqURI address:%s" % (mqURI))

    result = gConfig.getConfigurationTree("/Resources/MQServices", mqService, mqType, mqName)
    if not result["OK"] or not result["Value"]:
        return S_ERROR("Requested destination not found in the CS: %s::%s::%s" % (mqService, mqType, mqName))
    mqDestinationPath = None
    for path, value in result["Value"].items():
        if not value and path.endswith(mqName):
            mqDestinationPath = path

    # set-up internal parameter depending on the destination type
    tmp = mqDestinationPath.split("Queues")[0].split("Topics")
    servicePath = tmp[0]
    serviceDict = {}
    if len(tmp) > 1:
        serviceDict["Topic"] = mqName
    else:
        serviceDict["Queue"] = mqName

    result = gConfig.getOptionsDict(servicePath)
    if not result["OK"]:
        return result
    serviceDict.update(result["Value"])

    result = gConfig.getOptionsDict(mqDestinationPath)
    if not result["OK"]:
        return result
    serviceDict.update(result["Value"])
    return S_OK(serviceDict)


def getMQService(mqURI):
    return mqURI.split("::")[0]


def getDestinationType(mqURI):
    return mqURI.split("::")[1]


def getDestinationName(mqURI):
    return mqURI.split("::")[2]


def getDestinationAddress(mqURI):
    mqType, mqName = mqURI.split("::")[-2:]
    # We remove the trailing 's to change from Queues to Queue
    mqType = mqType.rstrip("s")
    return "/" + mqType.lower() + "/" + mqName


def generateDefaultCallback():
    """Function generates a default callback that can
        be used to handle the messages in the MQConsumer
        clients. It contains the internal queue (as closure)
        for the incoming messages. The queue can be accessed by the
        callback.get() method. The callback.get() method returns
        the first message or raise the exception Queue.Empty.
        e.g. myCallback = generateDefaultCallback()

            try:
               print myCallback.get()
            except Queue.Empty:
              pass

    Args:
      mqURI(str):Pseudo URI identifing MQ connection. It has the following format
                mqConnection::DestinationType::DestinationName
                e.g. blabla.cern.ch::Queues::MyQueue1
    Returns:
      object: callback function
    """
    msgQueue = Queue.Queue()

    def callback(headers, body):
        msgQueue.put(body)
        return S_OK()

    def get():
        return msgQueue.get(block=False)

    callback.get = get
    return callback



""" DIRAC FileCatalog Storage Element Manager base class for mix-in classes """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading

from DIRAC import S_ERROR


class SEManagerBase(object):
    def _refreshSEs(self, connection=False):
        """Refresh the SE cache"""

        return S_ERROR("To be implemented on derived class")

    def __init__(self, database=None):
        self.db = database
        if self.db:
            self.db.seNames = {}
            self.db.seids = {}
        self.lock = threading.Lock()
        self.seUpdatePeriod = 600

        # last time the cache was updated (epoch)
        self.lastUpdate = 0
        self._refreshSEs()

    def setUpdatePeriod(self, period):
        self.seUpdatePeriod = period

    def setDatabase(self, database):
        self.db = database
        self.db.seNames = {}
        self.db.seids = {}
        self._refreshSEs()



""" Plotting Service generates graphs according to the client specifications
    and data
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import hashlib

from DIRAC import S_OK, S_ERROR, rootPath, gConfig, gLogger
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.FrameworkSystem.Service.PlotCache import gPlotCache


def initializePlottingHandler(serviceInfo):

    # Get data location
    plottingSection = PathFinder.getServiceSection("Framework/Plotting")
    dataPath = gConfig.getValue("%s/DataLocation" % plottingSection, "data/graphs")
    dataPath = dataPath.strip()
    if "/" != dataPath[0]:
        dataPath = os.path.realpath("%s/%s" % (gConfig.getValue("/LocalSite/InstancePath", rootPath), dataPath))
    gLogger.info("Data will be written into %s" % dataPath)
    try:
        os.makedirs(dataPath)
    except Exception:
        pass
    try:
        testFile = "%s/plot__.test" % dataPath
        with open(testFile, "w"):
            pass
        os.unlink(testFile)
    except IOError:
        gLogger.fatal("Can't write to %s" % dataPath)
        return S_ERROR("Data location is not writable")

    gPlotCache.setPlotsLocation(dataPath)
    gMonitor.registerActivity("plotsDrawn", "Drawn plot images", "Plotting requests", "plots", gMonitor.OP_SUM)
    return S_OK()


class PlottingHandler(RequestHandler):
    def __calculatePlotHash(self, data, metadata, subplotMetadata):
        m = hashlib.md5()
        m.update(repr({"Data": data, "PlotMetadata": metadata, "SubplotMetadata": subplotMetadata}).encode())
        return m.hexdigest()

    types_generatePlot = [[dict, list], dict]

    def export_generatePlot(self, data, plotMetadata, subplotMetadata={}):
        """Create a plot according to the client specification and return its name"""
        plotHash = self.__calculatePlotHash(data, plotMetadata, subplotMetadata)
        result = gPlotCache.getPlot(plotHash, data, plotMetadata, subplotMetadata)
        if not result["OK"]:
            return result
        return S_OK(result["Value"]["plot"])

    def transfer_toClient(self, fileId, token, fileHelper):
        """
        Get graphs data
        """
        retVal = gPlotCache.getPlotData(fileId)
        if not retVal["OK"]:
            return retVal
        retVal = fileHelper.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        fileHelper.sendEOF()
        return S_OK()



"""
Helper class for configuring the monitoring service.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

########################################################################


class BaseType(object):

    """
    .. class:: BaseType

    :param str index: we use daily indexes for example:wmshistory_index-2015-10-09
    :param keyFields: The attributes what we monitor.
    :type keyFields: python:list
    :param monitoringFields: This is the value what we plot
    :type monitoringFields: python:list
    :param int dataToKeep: Data retention. We keep all data by default.
    :param dict mapping: We can specify the mapping of the documents. It is used during the creation of an index.
                         Note: If you do not want to be analysed a string, you have to set the mapping
    :param str period: We can configure the elasticsearch index name with a period.
                       By default we use daily indexes, but we can also use weekly, monthly, yearly indexes.
                       Or use no period at all.
                       Supported values: day, week, month, year, null

    """

    ########################################################################
    def __init__(self):
        """c'tor

        :param self: self reference
        """
        self.keyFields = []
        self.mapping = {}  # by default we do not have mapping
        self.timeType = {"properties": {"timestamp": {"type": "date"}}}  # we use timestamp for all monitoring types.
        self.period = "day"
        self.monitoringFields = ["Value"]
        self.index = None
        self.index = self._getIndex()

        # we only keep the last month of the data.
        self.dataToKeep = -1

    ########################################################################
    def checkType(self):
        """
        The mandatory fields has to be present
        """
        if not self.keyFields:
            raise Exception("keyFields has to be provided!")
        if not self.monitoringFields:
            raise Exception("monitoringFields has to be provided!")

    ########################################################################
    def _getIndex(self):
        """It returns and index based on the name of the type.
        For example: WMSMonitorType the type the index will be wmsmonitor
        """
        index = ""
        if self.index is None:
            fullName = self.__class__.__name__
            index = "%s-index" % fullName.lower()
        else:
            index = self.index
        return index

    ########################################################################
    def addMapping(self, mapping):
        """
        :param dict mapping: the mapping used by elasticsearch
        """
        self.mapping = self.timeType
        self.mapping["properties"].update(mapping)
